# AIS.R - Implementation of Annealed Importance Sampling.


# ANNEALED IMPORTANCE SAMPLING.  Performs one run of Annealed Importance
# Sampling (AIS), also known as Jarzynski's method, returning all the states 
# visited during the run, and all the ratios used for estimating ratios of 
# normalizing constants.  The run can be in either the `forward' or `reverse'
# direction with respect to the sequence of distributions specified by the
# given vector of values for the eta parameter (with a value of eta=0 being 
# assumed at the beginning of this sequence).  Markov chain updates for the
# distributions defined for all eta values (except eta=0 and the final eta)
# are performed by a specified `trans' function, which is passed a 'stepsize' 
# argument supplied for each eta value.  A state from the initial (eta=0)
# distribution must be provided as an argument for a forward run, or a state
# from the final distribution must be provided for a reverse run.
#
# The arguments are as follows:
#
#   lpr        function with arguments (x, eta) returning the log probability
#              (or density) of x under the distribution defined by eta, but
#              with the normalizing constant omitted.
#
#   eta        vector of parameter values for the distributions in the sequence,
#              except for the initial distribution (for which eta is zero).
#
#   trans      function with arguments (x, lpr, eta, stepsize) returning a new 
#              state produced by a Markov chain transition from x, satisfying
#              detailed balance with respect to the distribution defined by lpr
#              and eta.  The stepsize argument may modify the transition.
#           
#   stepsize   vector of stepsize arguments to trans, to be used for each
#              distribution in the sequence (except the initial and final 
#              distributions).  Extended or truncated to the length of eta 
#              minus one by repetition (if necessary).
#
#   initial    a vector which is sampled from the initial (eta=0) state 
#              distribution; must be provided for a forward run.
#
#   final      a vector which is sampled from the final state distribution;
#              must be provided for a reverse run.
#
# The value returned is a list with the following elements:
#
#   ratios     vector of the logs of the ratios whose product is an unbiased 
#              estimate of the ratio of the normalizing constant for the last 
#              distribution to the normalizing constant for the first.  For
#              a forward run, first=initial (eta=0) and last=final, whereas
#              for a reverse run this (and the order in 'ratios') is reversed.
#
#   states     matrix whose rows are the states found during the annealing run,
#              including the first state, supplied as the initial or final
#              argument.  Note that no state is produced using a transition 
#              with the last eta value (either the initial or final eta).
#
# To do both forward and reverse runs, that are otherwise the same, ais should
# be called twice, with the same values of lpr, eta, trans, and stepsize, but
# with one call specifying initial and the other specifying final.
#
# Note that if length(eta) is one, there are no intermediate distributions,
# and this function does simple importance sampling.

ais <- function (lpr, eta, trans, stepsize, initial=NULL, final=NULL)
{
  # Find number of distributions, not counting initial distribution with eta=0,
  # allocate space for ratios, and set up stepsizes.

  n <- length(eta)
  ratios <- rep(NA,n)
  stepsize <- rep(stepsize,length=n-1)

  # See if we're doing a forward or reverse AIS run, and initialize accordingly.

  if (is.null(initial) && is.null(final))
  { stop("Must specify a state from either the initial or final distribution")
  }
  if (is.null(final))
  { d <- length(initial)
    states <- matrix(NA,n,d)
    states[1,] <- initial
    prev.eta <- 0
    order <- 1:n
  }
  else if (is.null(initial))
  { d <- length(final)
    states <- matrix(NA,n,d)
    states[1,] <- final
    prev.eta <- eta[n]
    order <- (n-1):0
  }
  else
  { stop("Must not specify states from both initial and final distributions")
  }

  # Loop through the sequence of distributions (in either the forward or the
  # reverse order).

  for (i in 1:n)
  { 
    # Compute the log of the next ratio.

    if (order[i]==0) 
    { this.eta <- 0
    }
    else
    { this.eta <- eta[order[i]]
    }

    ratios[i] <- lpr(states[i,],this.eta) - lpr(states[i,],prev.eta)

    if (i<n && i>0)
    {
      # Simulate the next transition.

      states[i+1,] <- trans(states[i,],lpr,this.eta,stepsize[order[i]])

      # Go on to the next distribution.

      prev.eta <- this.eta
    }
  }
  
  list (ratios=ratios, states=states)
}
