# EST.R - Find final estimates of ratios of normalizing constants.


# FIND LIS/AIS ESTIMATES FOR THE RATIO OF NORMALIZING CONSTANTS.  The argument
# is a list whose elements are lists of results as returned by lis or ais.  The
# value returned is a vector containing the logs of the estimates for the ratio
# of normalizing constants found by each of these LIS or AIS runs.

loge <- function (results)
{
  n.runs <- length(results)
  s <- rep(NA,n.runs)

  for (i in 1:length(results))
  { s[i] <- sum(results[[i]]$ratios)
  }

  s
}


# FIND SIMPLE LIS/AIS ESTIMATES FOR THE RATIO OF NORMALIZING CONSTANTS.  The 
# argument is a vector whose elements are logs of estimated ratios from lis 
# or ais.  The result is a 2x2 matrix containing two estimates (first column)
# and their standard errors (second column) in its two rows.  The estimates
# are for the ratio of normalizing constants, and for the log of this ratio
# (which is simply the log of the estimate for the ratio).  Note that the 
# estimates for ratio may underflow, but this will probably not be a problem 
# for the log of the estimate.  The standard errors will be unreliable if the 
# standard error for the estimate of the log ratio is greater than about 0.2.

simple.est <- function (e)
{ 
  n <- length(e)
  m <- max(e)
  e <- exp(e-m)

  matrix(c(mean(e)*exp(m),       log(mean(e))+m, 
           sd(e)*exp(m)/sqrt(n), sd(e)/mean(e)/sqrt(n)),
         nrow=2, ncol=2, 
         dimnames = list (c("ratio","log ratio"), c("estimate","std.err")))
}


# FIND ADJUSTED LIS/AIS ESTIMATES FOR THE RATIO OF NORMALIZING CONSTANTS.  
# Like simple.est, except that the estimates are adjusted on the assumption
# that the logs of the individual estimates passed as the argument are 
# normally distributed.  This may or may not improve results in practice.

adjusted.est <- function (e)
{ 
  n <- length(e)

  mu <- mean(e)
  ss <- var(e)

  
  m <- max(e)
  e <- exp(e-m)

  matrix(c(exp(mu+ss/2),                          mu+ss/2,
           exp(mu+ss/2)*sqrt(ss/n+2*ss^2/(n-1)), sqrt(ss/n+2*ss^2/(n-1))),
         nrow=2, ncol=2, 
         dimnames = list (c("ratio","log ratio"), c("estimate","std.err")))
}


# FIND BRIDGED LIS/AIS ESTIMATES FOR THE RATIO OF NORMALIZING CONSTANTS.  The 
# first two arguments are vectors whose elements are logs of estimated ratios 
# from forward and reverse calls of lis or ais (with everything the same for the
# two calls except forward vs. reverse).  If the "use" argument is non-null,
# only the first "use" elements of these vectors are used.  The remaining 
# arguments specify an initial estimate for the log of the ratio of normalizing 
# constants, the number of iterations to improve this estimate when using the
# optimal form of bridge, and the function used to find lower-level estimates
# (either simple.est or adjusted.est).  The geometric bridge and simple.est are
# the defaults.  The result is a 2x2 matrix containing two estimates (first 
# column) and their standard errors (second column) in its two rows.  The 
# estimates are for the ratio of normalizing constants, and for the log of this
# ratio (which is simply the log of the estimate for the ratio).  Note that the 
# estimates for ratio may underflow, but this will probably not be a problem 
# for the log of the estimate.  The standard errors will be unreliable if the 
# standard error for the estimate of the log ratio is greater than about 0.2.

bridged.est <- function (fe, re, use=NULL, log.iest=NULL, n.iter=0, 
                         est=simple.est)
{ 
  if (!is.null(use))
  { use <- ceiling(use)
    fe <- fe[1:use]
    re <- re[1:use]
  }

  N.0 <- length(fe)
  N.1 <- length(re)

  if (is.null(log.iest))  # geometric bridge
  { 
    ne <- fe/2
    de <- re/2
    nr <- est(ne)
    dr <- est(de)
    br <- nr
    br[2,1] <- nr[2,1] - dr[2,1]
    br[2,2] <- sqrt (nr[2,2]^2 + dr[2,2]^2)
  }

  else  # optimal form of bridge
  { 
    logr <- log.iest
    for (i in 0:n.iter)
    { ne <- - log.add.exp (logr+log(N.0/N.1)-fe, 0)
      de <- - log.add.exp (logr+log(N.0/N.1), -re)
      nr <- est(ne)
      dr <- est(de)
      br <- nr
      br[2,1] <- nr[2,1] - dr[2,1]
      br[2,2] <- sqrt (nr[2,2]^2 + dr[2,2]^2)
      logr <- br[2,1]
    }
  }

  br[1,1] <- exp(br[2,1])
  br[1,2] <- br[2,2]*br[1,1]

  br
}


# REVERSE ESTIMATES.  Takes estimates for 1/r and converts them to estimates
# for r.

rev.est <- function (e)
{
  e[2,1] <- -e[2,1]
  e[1,1] <- exp(e[2,1])
  e[1,2] <- e[1,1]*e[2,2]

  e
}
