# LIS.R - Implementation of Linked Importance Sampling.


# LINKED IMPORTANCE SAMPLING.  Performs one run of Linked Importance Sampling, 
# (LIS) returning all the states visited by each chain simulated, and 
# all the ratios used for estimating ratios of normalizing constants.  The
# run can be in either the `forward' or `reverse' direction with respect to the
# sequence of distributions specified by the given vector of values for the 
# eta parameter (with a value of eta=0 being assumed at the beginning of this
# sequence).  Markov chain updates for the distributions defined for all eta 
# values (except eta=0 and the final eta) are performed by a specified 
# `trans' function, which is passed a 'stepsize' argument supplied for each 
# eta value.  A chain of states from either the initial (eta=0) distribution 
# (for a forward run), or from the final distribution (for a reverse run) must 
# be provided as an argument.
#
# The arguments are as follows:
#
#   lpr        function with arguments (x, eta) returning the log probability
#              (or density) of x under the distribution defined by eta, but
#              with the normalizing constant omitted.
#
#   eta        vector of parameter values for the distributions in the sequence,
#              except for the initial distribution (for which eta is zero).
#
#   trans      function with arguments (x, lpr, eta, stepsize) returning a new 
#              state produced by a Markov chain transition from x, satisfying
#              detailed balance with respect to the distribution defined by lpr
#              and eta.  The stepsize argument may modify the transition.
#           
#   stepsize   vector of stepsize arguments to trans, to be used for each
#              distribution in the sequence.  Extended or truncated to the 
#              length of eta plus one by repetition (if necessary).  The first 
#              element is not used for forward runs; the last is not used for 
#              reverse runs.
#
#   n.trans    vector containing the numbers of transitions in the chains for
#              each distribution.  Extended or truncated to the length of eta 
#              plus one by repetition (if necessary).  The first element is not
#              used for forward runs; the last is not used for reverse runs.
#
#   initial    a matrix whose rows contain states sampled from the initial
#              distribution, either independently, or by a Markov chain that
#              leaves the initial distribution invariant, started with a state 
#              drawn from the initial distribution.  Must be provided for a
#              forward run.
#
#   final      a matrix whose rows contain states sampled from the final
#              distribution, either independently, or by a Markov chain that
#              leaves the final distribution invariant, started with a state 
#              drawn from the final distribution.  Must be provided for a
#              reverse run.
#
#   rho0, rho1 powers to be used for the power form of the bridge.  Both
#              default to 1/2, if neither they nor iratios (below) is specified.
#              The power form of the bridge distribution between p0 and p1 is 
#              p0(x)^rho0 * p1(x)^rho1.
#
#   iratios    vector of guesses at log ratios to be used for the optimal form 
#              of the bridge, extended or truncated to the length of eta by 
#              repetition if necessary.  Not allowed if rho0 or rho1 specified. 
#
# The value returned is a list with the following elements:
#
#   ratios     vector of the logs of the ratios whose product is an unbiased 
#              estimate of the ratio of the normalizing constant for the last 
#              distribution to the normalizing constant for the first.  For
#              a forward run, first=initial (eta=0) and last=final, whereas
#              for a reverse run this (and the order in 'ratios') is reversed.
#              Note that the vector produced by a forward run is suitable
#              for use as the iratios argument in a subsequent call. 
#
#   chains     list in which the i'th element is a matrix whose rows are the 
#              states of the i'th chain, which satisfies detailed balance with
#              respect to the i'th distribution, not counting the first chain
#              (whose states were passed as the initial or final argument).
#              Note that these chains are not necessarily in equilibrium.  
#
#   links      vector in which the i'th element is the index (from 1) of the 
#              link state in the i'th chain stored in 'chains'.
#
# To do both a forward and a reverse run, with reversal the only difference, 
# the lis function should be called twice, with the same values for lpr, eta,
# trans, stepsize, n.trans, rho0, rho1, and iratios, but with one call 
# specifying initial and the other specifying final.

lis <- function (lpr, eta, trans, stepsize, n.trans, initial=NULL, final=NULL,
                 rho0=NULL, rho1=NULL, iratios=NULL)
{
  # Check arguments for validity.

  if (any(n.trans<0))
  { stop("An element of n.trans is less than zero")
  }

  if ((!is.null(rho0) || !is.null(rho1)) && !is.null(iratios))
  { stop("Can't specify rho0 or rho1 and also iratios")
  }

  # Find number of distributions, not counting initial distribution with eta=0.

  n <- length(eta)

  # See if we're doing a forward or reverse LIS run, and initialize accordingly.

  if (is.null(initial) && is.null(final))
  { stop("Must specify a state from either the initial or final distribution")
  }
  if (is.null(final))
  { d <- ncol(initial)
    prev.n.trans <- nrow(initial)-1
    prev.chain <- initial
    prev.eta <- 0
    order <- 1:n
  }
  else if (is.null(initial))
  { d <- ncol(final)
    prev.n.trans <- nrow(final)-1
    prev.chain <- final
    prev.eta <- eta[n]
    order <- (n-1):0
  }
  else
  { stop("Must not specify states from both initial and final distributions")
  }

  # Set up vectors of stepsizes and numbers of transitions.  Note that these
  # are indexed by the index of the distribution (0 to n) plus one.

  stepsize <- rep(stepsize,length=n+1)
  n.trans <- rep(n.trans,length=n+1)

  # Set default values for rho0 and rho1 if needed.

  if (is.null(iratios))
  { if (is.null(rho0)) rho0 <- 1/2
    if (is.null(rho1)) rho1 <- 1/2
  }

  if (!is.null(iratios))
  { iratios <- rep(iratios,length=n)
    if (is.null(initial))
    { iratios <- -iratios
    }
  }

  ratios <- rep(NA,n)
  links <- rep(NA,n)
  chains <- list()

  # Loop through the sequence of distributions.

  for (i in 1:n)
  { 
    if (order[i]==0) 
    { this.eta <- 0
    }
    else
    { this.eta <- eta[order[i]]
    }

    this.stepsize <- stepsize[1+order[i]]
    this.n.trans <- n.trans[1+order[i]]

    # Allocate space for new chain.

    chain <- matrix(NA,this.n.trans+1,d)

    # Randomly decide how many of this.n.trans transitions come before the 
    # link state.

    n.before <- sample(0:this.n.trans,1)

    # Select the link state from the previous chain, and compute the numerator
    # of the ratio.

    lr <- rep(NA,nrow(prev.chain))
    for (j in 1:nrow(prev.chain))
    { lpr0 <- lpr(prev.chain[j,],prev.eta)
      lpr1 <- lpr(prev.chain[j,],this.eta)
      if (is.null(iratios))
      { lr[j] <- rho1*lpr1 + (rho0-1)*lpr0
      }
      else
      { lr[j] <- lpr1 - 
          log.add.exp (iratios[i]+log(prev.n.trans/this.n.trans)+lpr0, lpr1)
      }
    }
    lr.sum <- log.sum.exp(lr)
    k <- sample(1:nrow(prev.chain),1,prob=exp(lr-lr.sum))
    lnk <- prev.chain[k,]

    chain[n.before+1,] <- lnk
    links[i] <- n.before+1
    log.numerator <- lr.sum - log(nrow(prev.chain))

    # Simulate the states before the link state.

    if (n.before>0)
    { for (j in n.before:1)
      { chain[j,] <- trans(chain[j+1,],lpr,this.eta,this.stepsize)
      }
    }

    # Simulate the states after the link state.

    if (n.before<this.n.trans)
    { for (j in (n.before+2):(this.n.trans+1))
      { chain[j,] <- trans(chain[j-1,],lpr,this.eta,this.stepsize)
      }
    }

    # Compute the denominator of the ratio.

    lr <- rep(NA,nrow(chain))
    for (j in 1:nrow(chain))
    { lpr0 <- lpr(chain[j,],prev.eta)
      lpr1 <- lpr(chain[j,],this.eta)
      if (is.null(iratios))
      { lr[j] <- rho0*lpr0 + (rho1-1)*lpr1
      }
      else
      { lr[j] <- lpr0 - 
          log.add.exp (iratios[i]+log(prev.n.trans/this.n.trans)+lpr0, lpr1)
      }
    }
    lr.sum <- log.sum.exp(lr)
    log.denominator <- lr.sum - log(nrow(chain))

    ratios[i] <- log.numerator - log.denominator

    # Go on to the next distribution.

    chains[[i]] <- chain
    prev.chain <- chain
    prev.eta <- this.eta
    prev.n.trans <- this.n.trans
  }
  
  list (ratios=ratios, chains=chains, links=links)
}
