# TEST.R - Test the various methods on a given distribution sequence.


# RUN TESTS OF AIS AND LIS.  Arguments are as follows:
#
#   true      True value of the ratio of normalizing constants.
#
#   rep       Number of repetitions of AIS and LIS estimates used to evaluate 
#             the properties of these estimates.
#
#   run.ais   Function to find an AIS estimate, or NULL.  The function's
#             arguments are the number of runs, the number of distributions
#             used, not counting the one with eta=0, and a flag set to TRUE
#             for a reverse run.
#
#   M.ais     Number of runs to do of AIS.  Only M.ais/2 of these are used when
#             computing bridged AIS estimates, so that the computation time for
#             these will be similar to that for unbridged estimates.
#
#   n.ais     Number of distributions used with AIS, not counting the one with 
#             eta=0. 
#
#   run.lis   Function to find an LIS estimate, or NULL.  The function's
#             arguments are the number of runs, the number of distributions
#             used, not counting the one with eta=0, the number of transitions
#             for each distribution, and a flag set to TRUE for a reverse run.
#
#   M.lis     Number of runs to do of LIS.  Only M.lis/2 of these are used when
#             computing bridged LIS estimates, so that the computation time for
#             these will be similar to that for unbridged estimates.
#
#   n.lis     Number of distributions used with LIS, other than the one with 
#             eta=0. 
#
#   n.trans   Number of transitions done with LIS for each distribution.  The
#             number of points sampled independently to start LIS is n.trans+1.
#
#   seed      Random seed to use (set once at the beginning of the whole set
#             of tests).
#
# The value returned is a matrix with labelled rows and columns giving the
# properties of various AIS and LIS methods.

test <- function (true, rep, run.ais=NULL, M.ais, n.ais, 
                             run.lis=NULL, M.lis, n.lis, n.trans, 
                             seed=1)
{ 
  results <- function (e)
  { err <- e["ratio","estimate"] - true
    err.log <- e["log ratio","estimate"] - log(true)
    c (bias=err, mse=err^2, mse.se=(err^2)^2,
       gt.2.se=as.numeric(abs(err)>2*e["ratio","std.err"]),
       bias.log=err.log, mse.log=err.log^2, mse.log.se=(err.log^2)^2,
       gt.2.se.log=as.numeric(abs(err.log)>2*e["log ratio","std.err"])
    )
  }

  set.seed(seed)

  res <- 0

  for (i in 1:rep)
  {
    r <- NULL

    if (!is.null(run.ais))
    { r.ais.f <- run.ais (M.ais, n.ais, F)
      r.ais.r <- run.ais (M.ais, n.ais, T)
      r <- rbind (r,
       ais.f=results(simple.est(loge(r.ais.f))),
       ais.r=results(rev.est(simple.est(loge(r.ais.r)))),
       ais.bg=results(bridged.est(loge(r.ais.f),loge(r.ais.r),M.ais/2)),
       ais.bo=results(bridged.est(loge(r.ais.f),loge(r.ais.r),M.ais/2,0,10)) 
      )
    }

    if (!is.null(run.lis))
    { r.lis.fg <- run.lis (M.lis, n.lis, n.trans, F)
      r.lis.rg <- run.lis (M.lis, n.lis, n.trans, T)
      r.lis.fo <- run.lis (M.lis, n.lis, n.trans, F, iratios=log(true)/n.lis)
      r.lis.ro <- run.lis (M.lis, n.lis, n.trans, T, iratios=log(true)/n.lis)
      r <- rbind (r,
       lis.fg=results(simple.est(loge(r.lis.fg))),
       lis.rg=results(rev.est(simple.est(loge(r.lis.rg)))),
       lis.fo=results(simple.est(loge(r.lis.fo))),
       lis.ro=results(rev.est(simple.est(loge(r.lis.ro)))),
       lis.bgg=results(bridged.est(loge(r.lis.fg),loge(r.lis.rg),M.lis/2)),
       lis.bgo=results(bridged.est(loge(r.lis.fg),loge(r.lis.rg),M.lis/2,0,10)),
       lis.bog=results(bridged.est(loge(r.lis.fo),loge(r.lis.ro),M.lis/2)),
       lis.boo=results(bridged.est(loge(r.lis.fo),loge(r.lis.ro),M.lis/2,0,10))
      )
    }

    res <- r + res
  }

  res <- res/rep
  res[,"mse.se"] <- sqrt(res[,"mse.se"]-res[,"mse"]^2) / sqrt(rep)
  res[,"mse.log.se"] <- sqrt(res[,"mse.log.se"]-res[,"mse.log"]^2) / sqrt(rep)

  res
}
